"""Test class for setting up AMMs."""
from enum import Enum
from math import isclose
from typing import List, NewType, Tuple

from sympy import symbols

from Arbie import PoolValueError
from Arbie.Variables.token import Balance, Balances, Token, Tokens

x = symbols("x")


def get_value(values: Balances, token: Token) -> Balance:
    for v in values:
        if v.token == token:
            return v.value


class PoolType(Enum):
    uniswap = 0
    balancer = 1
    unknown = 99


class Pool(object):
    """Pool can create a arbitrary AMM that can be used for testing.

    The math can be found here https://balancer.finance/whitepaper/
    """

    def __init__(
        self,
        tokens: Tokens,
        balances: List[float],
        weights: List[float],
        pool_type: PoolType = PoolType.unknown,
        fee: float = 0,
        **kwargs,
    ):  # noqa: WPS221

        self.tokens = tokens
        self.balances = Balance.create(tokens, balances)
        self.weights = Balance.create(tokens, weights)
        self.fee = fee
        self.pool_type = pool_type

        self.address = None
        if "address" in kwargs:
            self.address = kwargs.get("address")

        if not isclose(sum(weights), 1, abs_tol=1e-3):  # noqa: WPS432
            raise PoolValueError(
                f"Weights are not normalized, sum is {sum(weights)} for pool {self.address}"
            )

        self._check_balances()

        if self.fee > 1 or self.fee < 0:
            raise PoolValueError(
                f"Fee: {self.fee}, should be between 0 and 1. Fee is {self.fee}, for pool {self.address}"
            )

    def __str__(self):
        return f"""
Pool(
  Tokens: {self.tokens},
  Balances: {self.balances},
  Weights: {self.weights},
  Fee: {self.fee}
  Address: {self.address})"""

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.address)

    def update_balances(self, balances: List[float]):
        self.balances = Balance.create(self.tokens, balances)
        self._check_balances()

    def get_weights(self, token_in: Token, token_out: Token) -> Tuple[float, float]:
        return (get_value(self.weights, token_in), get_value(self.weights, token_out))

    def get_balances(self, token_in: Token, token_out: Token) -> Tuple[float, float]:
        return (get_value(self.balances, token_in), get_value(self.balances, token_out))

    def spot_price(self, token_in: Token, token_out: Token) -> float:
        """Ratio between token_in and token_out.

        A ratio > 1 means that token_in is less valuable then token_out
        A ratio < 1 means that token_in is more valuable then token_out

        token_in / ratio = token_out

        If the ratio is 400 then it takes 400 token_in for 1 token_out
        """
        bi, bo = self.get_balances(token_in, token_out)
        wi, wo = self.get_weights(token_in, token_out)
        return (bi / wi) / (bo / wo)

    def in_given_out_expr(self, token_in: Token, token_out: Token):
        bi, bo = self.get_balances(token_in, token_out)
        wi, wo = self.get_weights(token_in, token_out)

        return bi * ((bo / (bo - x * (1 - self.fee))) ** (wo / wi) - 1)  # noqa: WPS221'

    def out_given_in_expr(self, token_in: Token, token_out: Token) -> float:
        bi, bo = self.get_balances(token_in, token_out)
        wi, wo = self.get_weights(token_in, token_out)

        return bo * (1 - (bi / (bi + x * (1 - self.fee))) ** (wi / wo))  # noqa: WPS221

    def in_given_out(self, token_in: Token, token_out: Token, amount: float) -> float:
        expr = self.in_given_out_expr(token_in, token_out)
        return expr.subs(x, amount)

    def out_given_in(self, token_in: Token, token_out: Token, amount: float) -> float:
        expr = self.out_given_in_expr(token_in, token_out)
        return expr.subs(x, amount)

    def _check_balances(self):
        for balance in self.balances:
            if isclose(balance.value, 0, abs_tol=1e-3):  # noqa: WPS432
                raise PoolValueError(f"Balance {balance} ~ 0 for pool {self.address}")


Pools = NewType("Pools", List[Pool])



"""Graph contains different trading graphs.

To create efficient Arbitrage algorithms we need graphs
suited for modelling relationships between Automated Market Makers.
"""
from typing import List

import networkx as nx

from Arbie.Variables.pool import Pool
from Arbie.Variables.token import Token


class Graph(object):
    """Base class for wrapping networkx graph."""

    def __len__(self):
        return len(self.graph)

    def __getitem__(self, key):
        return self.graph[key]

    def get_edges(self):
        return self.graph.edges

    def get_nodes(self):
        return self.graph.nodes

    def _add_edge(self, start_node: Token, end_node: Token, weight, pool):
        if start_node is end_node:
            return

        self.graph.add_edge(start_node, end_node, weight=weight, object=pool)


class TradingGraph(Graph):
    """A trading graph.

    Represents each possible trading path between all tokens.
    """

    def __init__(self, pools: List[Pool]):
        self.graph = nx.MultiDiGraph()
        self._create_graph(pools)

    def _create_graph(self, pools: List[Pool]):
        """Create graph from a set of Pools."""
        for pool in pools:
            self._add_nodes(pool.tokens)

        for pool in pools:  # noqa: WPS440,WPS441
            self._add_edges(pool)  # noqa: WPS441

    def _add_nodes(self, tokens: List[Token]):
        self.graph.add_nodes_from(tokens)

    def _add_edges(self, pool: Pool):  # noqa: WPS231
        for start_node in pool.tokens:
            for end_node in pool.tokens:
                if start_node == end_node:
                    continue
                try:
                    weight = pool.spot_price(start_node, end_node)
                except ZeroDivisionError:
                    raise ZeroDivisionError(
                        f"""Zero division error in
                    pool: {pool}
                    between token {repr(start_node)} and {repr(end_node)}
                    """
                    )

                self._add_edge(start_node, end_node, weight, pool)


class FilteredTradingGraph(Graph):
    """A Filtered Trading Graph.

    Represents the best trades between each token pair.
    The best trade is the one that has the highest ratio.
    """

    weight_key = "weight"
    pool_key = "object"

    def __init__(self, trading_graph: TradingGraph, min_edge_liquidity):
        self.min_edge_liquidity = min_edge_liquidity
        self.graph = nx.DiGraph()
        self._filter_graph(trading_graph)

    def _check_node_liquidity(self, balance, node: Token) -> bool:
        return balance * node.price < self.min_edge_liquidity

    def _check_liquidity(self, start_node: Token, end_node: Token, pool: Pool) -> bool:
        t1, t2 = pool.get_balances(start_node, end_node)
        return self._check_node_liquidity(t1, start_node) or self._check_node_liquidity(
            t2, end_node
        )

    def _filter_graph(self, trading_graph: TradingGraph):  # noqa: WPS231
        self.graph.add_nodes_from(trading_graph.get_nodes())

        for (start_node, end_node, data) in trading_graph.get_edges().data():

            # Check if there is an edge between two nodes in the filtered graph
            edge_data = self.graph.get_edge_data(start_node, end_node)

            if self._check_liquidity(start_node, end_node, data[self.pool_key]):
                continue

            if edge_data is not None:
                if edge_data[self.weight_key] < data[self.weight_key]:
                    # The most valuable path is already added
                    continue

            self._add_edge(
                start_node, end_node, data[self.weight_key], data[self.pool_key]
            )



"""Base abstract class for pool contracts."""

from math import isclose
from typing import List

from Arbie import PoolValueError
from Arbie.async_helpers import async_map
from Arbie.Contracts.contract import Contract
from Arbie.Contracts.tokens import GenericToken
from Arbie.Variables import BigNumber, Pool, PoolType, Token


async def create_token(token_contract: GenericToken):
    return await token_contract.create_token()


class PoolContract(Contract):
    async def get_tokens(self) -> List[GenericToken]:
        raise NotImplementedError()

    async def get_balances(self) -> List[BigNumber]:
        raise NotImplementedError()

    async def get_weights(self) -> List[float]:
        raise NotImplementedError()

    async def get_fee(self) -> float:
        raise NotImplementedError()

    def get_type(self) -> PoolType:
        raise NotImplementedError()

    async def create_tokens(self) -> List[Token]:
        tokens = await self.get_tokens()
        return await async_map(create_token, tokens)

    async def create_pool(self) -> Pool:
        tokens = await self.create_tokens()
        if len(tokens) < 2:
            raise PoolValueError(
                f"Pool: {self.get_address()}, has insufficient tokens: {len(tokens)}"
            )
        balances_bg = await self.get_balances()
        balances = list(map((lambda bg: bg.to_number()), balances_bg))
        if isclose(sum(balances), 0, abs_tol=1e-3):  # noqa: WPS432
            raise PoolValueError(
                f"Pool: {self.get_address()}, balances {sum(balances)} == 0"
            )

        weights = await self.get_weights()
        if not isclose(sum(weights), 1, abs_tol=1e-3):  # noqa: WPS432
            raise PoolValueError(
                f"Pool: {self.get_address()}, weights {sum(weights)} != 1"
            )
        return Pool(
            tokens,
            balances,
            weights,
            self.get_type(),
            await self.get_fee(),
            address=self.get_address(),
        )



"""Unique exceptions for Arbie."""


class DeployContractError(Exception):
    """Raised when a contract can not be deployed."""


class IERC20TokenError(ValueError):
    """Raised when a IErc20 token is not Erc20 compliant."""


class PoolValueError(ValueError):
    """Raised when a Pool is not initialized properly.

    This can hapen when a pool dosn't have atleast 2 tokens.
    Or when the token weights dosn't add to one.
    """


class StateError(Exception):
    """Raised when there is an issue with the state."""


class TransactionError(RuntimeError):
    """Raised when a Transaction unexpectedly fails."""



"""Unittests of PathFinder."""
import asyncio
from unittest.mock import MagicMock

import pytest
import yaml

from Arbie import SettingsParser
from Arbie.Actions import Store
from Arbie.app import App


@pytest.fixture
def config_file():
    return """
    action_tree:
        actions:
            PathFinder:
                input:
                    weth: eth
                    min_liquidity: 4
                output:
                    cycles: found_cycles
            Arbitrage:
    """


@pytest.fixture
def store(pools, eth) -> Store:
    store = Store()
    store.add("pools", pools)
    future = asyncio.Future()
    future.set_result(eth)
    weth_mock = MagicMock()
    weth_mock.create_token.return_value = future
    store.add("eth", weth_mock)
    return store


class TestPathFinder(object):
    @pytest.mark.asyncio
    async def test_run(self, store, config_file, mocker):
        mocker.patch.object(SettingsParser, "setup_store", return_value=store)

        config = yaml.safe_load(config_file)
        app = App(config)
        assert len(app.action_tree.actions) == 2
        await app.run()
        assert len(store["found_cycles"]) == 5
        trades = app.store["filtered_trades"]
        assert trades[0].profit == pytest.approx(1.354642)



"""Utility functions for interacting with balancer."""

import logging
from typing import List

from Arbie import DeployContractError, IERC20TokenError
from Arbie.Contracts.contract import Contract, ContractFactory
from Arbie.Contracts.event_filter import EventFilter
from Arbie.Contracts.pool_contract import PoolContract
from Arbie.Contracts.tokens import GenericToken
from Arbie.Variables import BigNumber, PoolType

logger = logging.getLogger()


class BalancerPool(PoolContract):
    name = "BPool"
    protocol = "balancer"
    pool_type = PoolType.balancer

    def get_number_of_tokens(self):
        return self.contract.functions.getNumTokens().call()

    async def get_tokens(self) -> List[GenericToken]:
        token_addresses = await self._call_async(
            self.contract.functions.getCurrentTokens()
        )
        cf = ContractFactory(self.w3, GenericToken)
        return list(
            map(
                (
                    lambda a: cf.load_contract(
                        owner_address=self.owner_address, address=a
                    )
                ),
                token_addresses,
            )
        )

    async def get_balances(self) -> List[BigNumber]:
        tokens = await self.get_tokens()
        balances = []
        for token in tokens:
            b = self.contract.functions.getBalance(token.get_address()).call()
            try:
                decimals = await token.decimals()
            except Exception:
                raise IERC20TokenError("Bad token in balancer pool")
            balances.append(BigNumber.from_value(b, decimals))
        return balances

    async def get_weights(self) -> List[float]:
        tokens = await self.get_tokens()
        weights = list(
            map(
                (
                    lambda t: self.contract.functions.getNormalizedWeight(
                        t.get_address()
                    ).call()
                ),
                tokens,
            )
        )
        sum_of_weights = sum(weights)
        return list(map((lambda x: x / sum_of_weights), weights))

    async def get_fee(self) -> float:
        fee = await self._call_async(self.contract.functions.getSwapFee())
        return BigNumber.from_value(fee).to_number()

    def get_type(self) -> PoolType:
        return BalancerPool.pool_type

    def bind(self, address: str, balance: BigNumber, denorm_weight: int) -> bool:
        if denorm_weight < 1:
            raise ValueError("Weight should be larger than 1")
        eth_safe_weight = BigNumber(denorm_weight)
        transaction = self.contract.functions.bind(
            address, balance.value, eth_safe_weight.value
        )
        return self._transact_status(transaction)

    def finalize(self) -> bool:
        return self._transact_status(self.contract.functions.finalize())


class BalancerLoader(object):
    def __init__(self, factory: ContractFactory, address):
        self.factory = factory
        self.address = address

    def load_from_event(self, events):
        logger.info(f"Loading pool contracts in events {events}")
        return [
            self.factory.load_contract(
                owner_address=self.address, address=event.args.pool
            )
            for event in events
        ]


class BalancerFactory(Contract):
    name = "BFactory"
    protocol = "balancer"

    def __init__(self, w3, contract, **kwargs):
        self.cf = ContractFactory(w3, BalancerPool)
        super().__init__(w3, contract, **kwargs)

    async def setup_pool(
        self,
        tokens: List[GenericToken],
        weights: List[float],
        amounts: List[BigNumber],
        approve_owner=True,
    ) -> BalancerPool:
        pool = self.new_pool()

        for token, weight, amount in zip(tokens, weights, amounts):
            if approve_owner:
                await token.approve_owner()
            token.approve(pool.get_address(), amount)
            pool.bind(token.get_address(), amount, weight)

        pool.finalize()
        return pool

    def new_pool(self) -> BalancerPool:
        transaction = self.contract.functions.newBPool()
        status, address = self._transact_status_and_contract(transaction)

        if not status:
            raise DeployContractError("Failed to deploy BalancerPool.")

        return self.cf.load_contract(self.owner_address, address=address)

    async def all_pools(self, start=0, steps=100) -> List[BalancerPool]:
        last_block = self.w3.eth.blockNumber
        bf = EventFilter(
            self.contract.events.LOG_NEW_POOL,
            self._create_pools,
            start,
            last_block,
            steps,
        )
        return await bf.find_events()

    def _create_pools(self, new_pool_events):
        loader = BalancerLoader(self.cf, self.owner_address)
        return loader.load_from_event(new_pool_events)



"""Utility functions for interacting with smart contracts."""

import json
from enum import Enum
from typing import Tuple

from eth_account import Account
from pkg_resources import resource_string

from Arbie.async_helpers import run_async


class Network(Enum):
    mainnet = 0
    kovan = 1
    ropsten = 2


def _get_tx_params(w3, address, value=None, gas=None):
    """Get generic transaction parameters."""
    params = {
        "from": address,
        "nonce": w3.eth.getTransactionCount(address),
    }
    if value:
        params["value"] = value
    if gas:
        params["gas"] = gas
    return params


def _wait_for_tx(w3, tx_hash):
    return w3.eth.waitForTransactionReceipt(tx_hash, 120)  # noqa: WPS432


def transact(w3, address: str, transaction, value=None):
    """Transact a transaction and return transaction receipt."""
    tx_params = _get_tx_params(w3, address, value, 48814000)  # noqa: WPS432
    tx_hash = transaction.transact(tx_params)
    return _wait_for_tx(w3, tx_hash)


def signed_transaction(w3, user_account: Account, transaction, value, gas):
    tx_params = _get_tx_params(w3, user_account.address, value, gas)

    signed_txn = Account.sign_transaction(
        transaction.buildTransaction(tx_params), private_key=user_account.key
    )
    tx_hash = w3.eth.sendRawTransaction(signed_txn.rawTransaction)
    return _wait_for_tx(w3, tx_hash)


class Contract(object):
    """Base class for contracts."""

    def __init__(
        self, w3, contract, owner_address: str = None, user_account: Account = None
    ):
        self.w3 = w3
        self.owner_address = owner_address
        self.contract = contract
        self.user_account = user_account

    def get_address(self) -> str:
        return self.contract.address

    def set_owner_address(self, address):
        if self.user_account is None:
            self.owner_address = address
        else:
            raise PermissionError(
                "A user account is set, update that instead of owner address."
            )

    def set_account(self, account):
        self.user_account = account
        self.owner_address = None

    def _transact(self, transaction, value=None, gas=None):
        if self.user_account is None:
            return transact(self.w3, self.owner_address, transaction, value)
        return signed_transaction(self.w3, self.user_account, transaction, value, gas)

    def _transact_status(self, transaction, value=None, gas=None) -> bool:
        return self._transact(transaction, value, gas).status

    def _transact_status_and_contract(self, transaction) -> Tuple[bool, str]:
        tx_receipt = self._transact(transaction)
        return tx_receipt.status, tx_receipt.logs[1].address

    def _get_account(self):
        if self.user_account is None:
            return self.owner_address
        return self.user_account.address

    def _estimate_gas(self, transaction):
        tx_params = _get_tx_params(self.w3, self._get_account())
        built_transaction = transaction.buildTransaction(tx_params)
        # When estimateing gas te key gas can not be in the transaction
        built_transaction.pop("gas", None)
        return self.w3.eth.estimateGas(built_transaction)

    async def _call_async(self, function):
        return await run_async(function.call)


class ContractFactory(object):
    def __init__(self, w3, factory_class: Contract):
        self.w3 = w3

        if factory_class.name is None or factory_class.protocol is None:
            raise ValueError(f"{factory_class} dose not contain default parameters")
        self.factory_class = factory_class

    def load_contract(self, owner_address: str = None, **kwargs) -> Contract:
        """Load contract require address or network to be passed in kwargs."""
        address = self._read_address(**kwargs)
        contract = self._load_contract(address)
        user_account = self._read_account(**kwargs)
        return self.factory_class(
            self.w3,
            owner_address=owner_address,
            user_account=user_account,
            contract=contract,
        )

    def deploy_contract(self, owner_address: str, *args) -> Contract:
        contract_address = self._deploy_contract(owner_address, *args)
        contract = self._load_contract(contract_address)
        return self.factory_class(
            self.w3, owner_address=owner_address, contract=contract
        )

    def _deploy_contract(self, deploy_address: str, *args) -> str:
        """Deploy contract and pass on args to contract abi constructor."""
        contract = self.w3.eth.contract(
            abi=self._read_abi(), bytecode=self._read_bytecode()
        )

        transaction = contract.constructor(*args)
        tx_receipt = transact(self.w3, deploy_address, transaction)

        return tx_receipt.contractAddress

    def _read_resource(self, path: str, filename: str) -> str:
        if path is None:
            path = ""
        else:
            path = ".{0}".format(path)

        file_path = "Arbie.resources.contracts{0}".format(path)
        return resource_string(file_path, filename).decode("utf-8")

    def _get_address(self, network: Network):
        json_data = json.loads(self._read_resource(None, "contract_addresses.json"))

        return json_data[self.factory_class.protocol][self.factory_class.name][
            network.name
        ]

    def _read_abi(self):
        return self._read_json("abi")

    def _read_bytecode(self):
        return self._read_json("bytecode")

    def _read_json(self, key):
        filename = f"{self.factory_class.name}.json"
        json_data = self._read_resource(self.factory_class.protocol, filename)
        return json.loads(json_data)[key]

    def _read_address(self, **kwargs):
        if "address" in kwargs:
            return kwargs.get("address")
        if "network" in kwargs:
            return self._get_address(kwargs.get("network"))

        raise NameError("kwargs does not contain network or address")

    def _read_account(self, **kwargs):
        if "account" in kwargs:
            return kwargs.get("account")

    def _load_contract(self, address: str):
        return self.w3.eth.contract(address=address, abi=self._read_abi())



"""Utility functions for interacting with Uniswap."""
import asyncio
import logging
from typing import List, Tuple

from prometheus_async.aio import time

from Arbie import DeployContractError, IERC20TokenError
from Arbie.async_helpers import async_map, run_async
from Arbie.Contracts.contract import Contract, ContractFactory
from Arbie.Contracts.pool_contract import PoolContract
from Arbie.Contracts.tokens import GenericToken
from Arbie.prometheus import get_prometheus
from Arbie.Variables import BigNumber, PoolType

logger = logging.getLogger()

CREATE_PAIR = get_prometheus().summary(
    "uniswap_factory_create_pair_index", "Time for creating a pair"
)


async def create_reserve(result: Tuple[float, GenericToken]):
    (value, token) = result
    try:
        exp = await token.decimals()
    except Exception:
        raise IERC20TokenError("Token doesn't contain decimals.")
    return BigNumber.from_value(value, exp)


class UniswapPair(PoolContract):
    name = "UniswapV2Pair"
    protocol = "uniswap"

    fee = 0.003
    weight = 0.5

    pool_type = PoolType.uniswap

    def mint(self, address: str) -> bool:
        transaction = self.contract.functions.mint(address)
        return self._transact_status(transaction)

    async def get_token0(self) -> GenericToken:
        return await self._get_token(self.contract.functions.token0())

    async def get_token1(self) -> GenericToken:
        return await self._get_token(self.contract.functions.token1())

    async def get_tokens(self) -> List[GenericToken]:
        return await asyncio.gather(self.get_token0(), self.get_token1())

    async def get_balances(self) -> List[BigNumber]:
        result = await asyncio.gather(self._get_reserves(), self.get_tokens())
        ziped_result = list(zip(result[0], result[1]))
        return await async_map(create_reserve, ziped_result)

    async def get_fee(self) -> float:
        return self.fee

    def get_type(self) -> PoolType:
        return UniswapPair.pool_type

    async def get_weights(self) -> List[float]:
        return [self.weight, self.weight]

    async def _get_token(self, function) -> GenericToken:
        token_address = await self._call_async(function)
        cf = ContractFactory(self.w3, GenericToken)
        return cf.load_contract(self.owner_address, address=token_address)

    async def _get_reserves(self):
        return await self._call_async(self.contract.functions.getReserves())


class UniswapFactory(Contract):
    name = "UniswapV2Factory"
    protocol = "uniswap"

    def __init__(self, w3, contract, **kwargs):
        self.cf = ContractFactory(w3, UniswapPair)
        super().__init__(w3, contract, **kwargs)

    async def all_pairs_length(self) -> int:
        return await self._call_async(self.contract.functions.allPairsLength())

    async def get_pair_address(self, index) -> str:
        return await self._call_async(self.contract.functions.allPairs(index))

    async def all_pairs(self, sleep=0) -> List[UniswapPair]:
        number_of_pairs = await self.all_pairs_length()
        return await async_map(self._create_pair_index, range(number_of_pairs))

    async def create_pair(
        self, token_a: GenericToken, token_b: GenericToken
    ) -> UniswapPair:
        transaction = self.contract.functions.createPair(
            token_a.get_address(), token_b.get_address()
        )

        if not self._transact_status(transaction):
            raise DeployContractError("Failed to deploy UniswapPair")
        pair_nmb = await self.all_pairs_length()
        return await self._create_pair_index(pair_nmb - 1)

    async def setup_pair(
        self, tokens: List[GenericToken], amounts: List[BigNumber]
    ) -> UniswapPair:
        pair = await self.create_pair(tokens[0], tokens[1])

        for token, amount in zip(tokens, amounts):
            token.transfer(pair.get_address(), amount)
        try:
            pair.mint(self.owner_address)
        except Exception:
            raise ValueError(f"Failed to mint tokens {tokens[0]},{tokens[1]}")
        return pair

    @time(CREATE_PAIR)
    async def _create_pair_index(self, index) -> UniswapPair:
        address = await self.get_pair_address(index)
        logger.info(f"Creating pair number {index}")
        return await run_async(self._load_pair, address)

    def _load_pair(self, address):
        return self.cf.load_contract(self.owner_address, address=address)



"""Helper for using prometheus_client."""

import logging

from prometheus_client import Gauge, Summary, start_http_server


def singleton(class_):
    instances = {}

    def getinstance(*args, **kwargs):  # noqa: WPS430
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return getinstance


@singleton
class Prometheus(object):
    def __init__(self):
        try:
            start_http_server(8000)  # noqa: WPS432
        except OSError as e:
            logging.getLogger().info(e)

        self.metric_store = {}

    def gauge(self, name, description):
        return self._get_or_create(name, description, Gauge)

    def summary(self, name, description):
        return self._get_or_create(name, description, Summary)

    def _get_or_create(self, name, description, metric):
        valid_name = self._to_valid_name(name)
        if valid_name in self.metric_store:
            return self.metric_store[valid_name]
        self.metric_store[valid_name] = metric(valid_name, description)
        return self._get_or_create(valid_name, description, metric)

    def _to_valid_name(self, name):
        return name.replace(".", "_").lower()


def get_prometheus():
    return Prometheus()



"""Unittest of redis state."""
from unittest.mock import MagicMock

import pytest
from pytest_mock.plugin import MockerFixture

from Arbie import StateError
from Arbie.Actions import Action, RedisState, Store
from Arbie.Actions.action import Argument


@pytest.fixture
def local_store():
    return Store()


@pytest.fixture
def redis_store(redis_state):
    return Store(redis_state)


class MockAction(object):
    in_variable_name = "pools"
    in_constant_name = "amount"
    out_variable_name = "tokens"

    redis_item_key = "arbie.1.amount"
    redis_collection_key = "arbie.1.pools"
    redis_collection_item_key = "arbie.1.pools.0xA1cB32"

    in_settings = {in_variable_name: "uniswap", in_constant_name: 1337}
    out_settings = {out_variable_name: "good_tokens"}

    in_settings_parsed = {
        in_variable_name: Argument("uniswap"),
        in_constant_name: Argument(1337),
    }

    redis_in_settings_parsed = {
        in_variable_name: Argument("arbie.1.pools"),
        in_constant_name: Argument(1337),
    }

    out_settings_parsed = {out_variable_name: Argument("good_tokens")}


class TestAction(object):
    def test_get_input_output_settings(self, mocker: MockerFixture):
        mocker.patch(
            "Arbie.Actions.action.yaml.safe_load",
            return_value={
                Action.input_key: MockAction.in_settings,
                Action.output_key: MockAction.out_settings,
            },
        )
        action = Action()
        assert MockAction.in_settings_parsed == action.get_input_settings()
        assert MockAction.out_settings_parsed == action.get_output_settings()

    @pytest.mark.asyncio
    async def test_on_next(self):
        with pytest.raises(NotImplementedError):
            await Action().on_next(None)


@pytest.fixture
def mock_action():
    mock = MagicMock()
    mock.get_input_settings.return_value = MockAction.in_settings_parsed
    mock.get_output_settings.return_value = MockAction.out_settings_parsed
    return mock


@pytest.fixture
def mock_action_redis():
    mock = MagicMock()
    mock.get_input_settings.return_value = MockAction.redis_in_settings_parsed
    mock.get_output_settings.return_value = MockAction.out_settings_parsed
    return mock


class TestStore(object):
    def test_get_raises(self, local_store):
        with pytest.raises(KeyError):
            local_store.get("key")

    def test_get_add(self, local_store):
        local_store.add("key", 1)
        assert local_store.get("key") == 1

    def test_create_input_value_error(self, local_store, mock_action):
        with pytest.raises(ValueError):
            local_store.create_input(mock_action)

    def test_create_input(self, local_store, mock_action):
        local_store.add("uniswap", 1337)
        local_store.create_input(mock_action)

    def test_publish(self, local_store: Store):
        with pytest.raises(StateError):
            local_store.publish("random channel", "random message")

    def test_subscribe(self, local_store: Store):
        with pytest.raises(StateError):
            local_store.subscribe("random channel")

    def test_delete(self, local_store: Store):
        with pytest.raises(StateError):
            local_store.delete("random key")


class TestRedisStore(object):
    def test_get_raises(self, redis_store):
        with pytest.raises(KeyError):
            redis_store.get("key")

    def test_get_add(self, redis_store):
        redis_store.add("key", 1)
        assert redis_store.get("key") == 1

    def test_store_get(self, redis_state, mocker: MockerFixture):
        mock_get = mocker.patch.object(RedisState, "__getitem__")
        store = Store(redis_state)

        store[MockAction.redis_item_key]  # noqa: WPS428
        assert mock_get.called

    def test_store_add(self, redis_state, mocker: MockerFixture):
        mock_set = mocker.patch.object(RedisState, "__setitem__")
        store = Store(redis_state)

        store.add(MockAction.redis_collection_key, None)
        assert mock_set.called

    def test_create_input_value_error(self, redis_store, mock_action):
        with pytest.raises(ValueError):
            redis_store.create_input(mock_action)

    def test_create_input(self, redis_store, mock_action):
        redis_store.add("uniswap", 1337)
        redis_store.create_input(mock_action)

    def test_create_input_from_redis(self, redis_store, mock_action_redis, mocker):
        mocker.patch.object(RedisState, "__contains__", return_value=True)
        redis_store.create_input(mock_action_redis)



"""Unittest of async_helpers.py."""
import asyncio

import pytest

from Arbie.async_helpers import async_map


async def pass_through(some_value):
    await asyncio.sleep(0)
    return some_value


@pytest.mark.asyncio
async def test_async_map():
    res = await async_map(pass_through, range(3))
    assert res == [0, 1, 2]


@pytest.mark.asyncio
async def test_async_map_chunk():
    res = await async_map(pass_through, range(3), 2)
    assert res == [0, 1, 2]



"""Unittests of PoolContracts."""

from unittest.mock import MagicMock

import pytest

from Arbie.Contracts.pool_contract import PoolContract

pytestmark = pytest.mark.asyncio


async def should_raise(promise):
    with pytest.raises(NotImplementedError):
        await promise


class TestPoolContract(object):
    async def test_should_throw(self):
        pc = PoolContract(MagicMock(), None, MagicMock())

        await should_raise(pc.create_tokens())

        await should_raise(pc.get_balances())

        await should_raise(pc.get_weights())

        await should_raise(pc.get_fee())

        with pytest.raises(NotImplementedError):
            pc.get_type()



"""Basic token representation."""
from typing import List, NewType


class Token(object):
    """Token is a light weight representation of a ERC20 token."""

    def __init__(self, name: str, address: str, price: float = None):
        self.name = name
        self.price = price
        self.address = address

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Token(Name: {self.name}, Price: {self.price}, Address: {self.address})"  # noqa: WPS221

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __hash__(self):
        return hash(self.address)


class Balance(object):
    """Balance of token."""

    def __init__(self, token: Token, value: float):
        self.token = token
        self.value = value

    def __str__(self):
        return f"Balance(Token: {self.token}, Value: {self.value})"

    def __repr__(self):
        return self.__str__()

    @classmethod
    def create(cls, tokens: List[Token], values: List[float]) -> List[object]:
        if len(tokens) != len(values):
            raise ValueError("All inputs must be of same length.")

        balances = []
        for token, value in zip(tokens, values):
            balances.append(cls(token, value))
        return balances


Balances = NewType("Balances", List[Balance])
Tokens = NewType("Tokens", List[Token])



"""Unittest of Token and Balance."""
import pytest

from Arbie.Variables import Balance, Token

dai_string = "dai"

dai = Token(dai_string, "aaaa")
eth = Token("eth", "bbbb")
tokens = [dai, eth]


class TestToken(object):
    def test_str(self):
        dai_str = dai.__str__()

        assert dai_string in dai_str

    def test_repr(self):
        dai_repr = dai.__repr__()

        assert dai_string in dai_repr
        assert "aaaa" in dai_repr


class TestBalance(object):
    def test_create(self):
        Balance.create(tokens, [1, 2])

    def test_create_wrong_len(self):
        with pytest.raises(ValueError):
            Balance.create(tokens, [1, 2, 3])
        with pytest.raises(ValueError):
            Balance.create(tokens, [1])

    def test_str_and_repr(self):
        bal = Balance(dai, 2.5)
        bal_str = bal.__str__()
        bal_repr = bal.__repr__()

        assert bal_str == bal_repr
        assert dai_string in bal_str
        assert "2.5" in bal_str



"""Variable is a module with basic building blocks."""

from Arbie.Variables.big_number import BigNumber  # noqa: F401
from Arbie.Variables.pool import Pool, Pools, PoolType  # noqa: F401
from Arbie.Variables.token import Balance, Balances, Token, Tokens  # noqa: F401
from Arbie.Variables.trades import Trade  # noqa: F401



"""Pool finder is responsible for finding all pools."""
import asyncio
import logging
from math import isclose
from typing import List, Tuple

from prometheus_async.aio import time

from Arbie import IERC20TokenError, PoolValueError
from Arbie.Actions.action import Action
from Arbie.async_helpers import async_map
from Arbie.Contracts import BalancerFactory, GenericToken, UniswapFactory, UniswapPair
from Arbie.Contracts.pool_contract import PoolContract
from Arbie.prometheus import get_prometheus
from Arbie.Variables import Pools, Token, Tokens

logger = logging.getLogger()

CREATE_TOKEN_TIME = get_prometheus().summary(
    "pool_finder_create_and_check_token", "Time for creating and checking a token"
)


def check_and_get_price(balances) -> Tuple[bool, float]:
    for balance in balances:
        if isclose(balance.value, 0, abs_tol=1e-3):  # noqa: WPS432
            return True, 0
    return False, balances[0] / balances[1]


class TokenFinder(object):
    def __init__(self, uoa):
        self.uoa = uoa

    async def create_token(self, pair: UniswapPair, price):
        tokens = await pair.get_tokens()
        t0 = tokens[0]
        t1 = tokens[1]
        if self.uoa.address == t0.get_address():
            return await t1.create_token(price)
        if self.uoa.address == t1.get_address():
            return await t0.create_token(1 / price)

    @time(CREATE_TOKEN_TIME)
    async def create_and_check_token(self, pair):
        balances = None
        try:
            balances = await pair.get_balances()
        except IERC20TokenError:
            return None

        is_zero, price = check_and_get_price(balances)
        if is_zero:
            return None

        token = await self.create_token(pair, price)
        logger.info(f"Finished creating token from {pair.get_address()}")
        return token

    async def create_tokens(self, uniswap_pairs: List[UniswapPair]) -> List[Token]:
        tokens = await async_map(self.create_and_check_token, uniswap_pairs)
        tokens.append(self.uoa)
        token_set = set(tokens)
        token_set.discard(None)
        return list(token_set)


async def create_and_filter_pools(
    pool_contracts: List[PoolContract], tokens: List[Tokens]
) -> Pools:
    pools = []
    for contract in pool_contracts:
        logger.info(
            f"Filtering contract {pool_contracts.index(contract)} of {len(pool_contracts)}"
        )
        try:
            pool = await contract.create_pool()
        except IERC20TokenError as e:
            logger.warning(f"Failed to create pool, bad token. {e}")
            continue
        except PoolValueError as e:
            logger.warning(f"Failed to create pool, bad pool. {e}")
            continue
        for token in pool.tokens:
            try:
                index = tokens.index(token)
            # Token not found in pool
            except ValueError:
                continue
            if token in tokens:
                token.price = tokens[index].price
        pools.append(pool)
    return pools


class PoolFinder(Action):
    """PoolFinder is responsible for finding pools.

    [Settings]
    input:
        weth: weth
        uniswap_factory: uniswap_factory
        balancer_factory: balancer_factory
        balancer_start: 9562480
    output:
        pools: all_pools
        tokens: all_tokens
    """

    async def on_next(self, data):
        weth = await data.weth().create_token(1)

        uni_coro = self._get_pairs_and_tokens(data.uniswap_factory(), weth)

        pools, tokens = await self._get_pools_and_tokens(
            *await self._get_results(uni_coro),
        )

        data.pools(pools)
        data.tokens(tokens)

    async def _get_pools_and_tokens(self, pair_contracts, tokens):
        result = await asyncio.gather(
            create_and_filter_pools(pair_contracts, tokens),
        )
        return result[0], tokens

    async def _get_results(
        self, uni_coro
    ) -> Tuple[List[BalancerFactory], List[UniswapFactory], List[GenericToken]]:
        results = await asyncio.gather(uni_coro)
        pairs = results[0][0]
        tokens = results[0][1]
        return pairs, tokens

    async def _get_pairs_and_tokens(self, factory: UniswapFactory, weth):
        pairs = await factory.all_pairs()
        logging.getLogger().info("Found all uniswap pairs, filtering tokens.")
        tokens = await TokenFinder(weth).create_tokens(pairs)
        return pairs, tokens



"""ActionTree is a set of Actions."""

import asyncio
import inspect
import logging
import sys
from typing import Dict, List, Tuple

from Arbie.Actions.action import Action, Store
from Arbie.prometheus import get_prometheus

RUN_TIME = get_prometheus().summary("arbie_run", "Time spent running actions")


def is_class_action(member):
    return inspect.isclass(member) and Action in member.__bases__  # noqa: WPS609


def get_all_actions(extra_actions=None) -> List[Tuple[str, type(Action)]]:
    actions = inspect.getmembers(sys.modules["Arbie.Actions"], is_class_action)
    if extra_actions is not None:
        return actions + extra_actions
    return actions


def create_action(name, config, extra_actions):
    for name_cls, action_cls in get_all_actions(extra_actions):
        if name_cls == name:
            return action_cls(config)
    raise ValueError(f"Action: {name} not found.")


class ActionTree(object):
    def __init__(self, store: Store):
        self.store = store
        self.actions = []
        self.channels = []
        self.is_stopped = False

    def register_event(self, event_channel):
        self.channels.append(self.store.subscribe(event_channel))

    @classmethod
    def create(cls, action_configs: Dict, store: Store, extra_actions=None):
        tree = cls(store)
        for key, config in action_configs.items():
            action = create_action(key, config, extra_actions)
            tree.add_action(action)
        return tree

    def add_action(self, action: Action):
        self.actions.append(action)

    async def run(self):
        with RUN_TIME.time():
            self.is_stopped = False
            if self.channels:
                await self._run_continuous()
            else:
                await self._run_once()

    def stop(self):
        self.is_stopped = True

    async def _run_continuous(self):
        while not self.is_stopped:
            new_messages = self._get_messages()
            if new_messages:
                logging.getLogger().info(f"New messages {new_messages}")
                await self._run_once()
            else:
                await asyncio.sleep(0.1)

    async def _run_once(self):
        for action in self.actions:
            data = self.store.create_input(action)
            action_name = action.__class__.__name__
            with get_prometheus().summary(
                f"{action_name.lower()}_time",
                f"Time taken to process action {action_name}",
            ).time():
                await action.on_next(data)

    def _get_messages(self):
        messages = list(map(lambda c: c.get_message(True), self.channels))
        # Remove None
        return [message for message in messages if message]



"""Packaging logic for arbie."""
import os

from setuptools import find_packages, setup


def read(path):
    with open(path) as req:
        return req.read().splitlines()


def package_files(directory, file_filter):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            if any(substring in filename for substring in file_filter):
                paths.append(os.path.join("..", path, filename))
    return paths


extra_files = package_files("Arbie/resources", [".json"])

setup(
    packages=find_packages(exclude=["tests", "unit", "system"]),
    package_data={"": extra_files},
    install_requires=read("pip/requirements.txt"),
    tests_require=read("pip/requirements-dev.txt"),
    entry_points={"console_scripts": ["Arbie = Arbie.__main__:main"]},
)
